﻿using System;
using System.Collections.Generic;
using System.Dynamic;
using System.Linq;

namespace DeaneBarker.Kiwi
{
    public class DocumentModel<T> where T : KiwiDocument
    {
        protected T doc;

        public string Content => doc.Content;
        public string Title => doc.Title;
        public int WordCount => doc.WordCount;

        public SectionModel Start { get; private set; }
        public SectionModel End { get; private set; }

        public ExpandoObject Sections { get; private set; }
        public ExpandoObject Section { get; private set; }
        public ExpandoObject Meta { get; private set; }

        public DocumentModel(T doc)
        {
            this.doc = doc;

            if(doc.HasSection("_start"))
            {
                Start = new SectionModel(doc.GetSection("_start"));
            }

            if (doc.HasSection("_end"))
            {
                End = new SectionModel(doc.GetSection("_end"));
            }

            Meta = new ExpandoObject();
            foreach(var meta in doc.Meta)
            {
                ((IDictionary<string, object>)Meta)[NormalizeMetaKey(meta.Key)] = meta.Value;
            }

            Sections = new ExpandoObject();
            ((IDictionary<string, object>)Sections)["all"] = doc.FilteredSections.Select(s => new SectionModel(s));
            Section = new ExpandoObject();
            foreach (var distinctSection in doc.Sections.GroupBy(s => s.Name))
            {
                ((IDictionary<string, object>)Sections)[NormalizeMetaKey(distinctSection.Key)] = distinctSection.Select(s => new SectionModel(s));
                ((IDictionary<string, object>)Section)[NormalizeMetaKey(distinctSection.Key)] = new SectionModel(distinctSection.First());
            }

        }

        public IEnumerable<SectionModel> this[string key]
        {
            get
            {
                return doc.Sections.Where(s => s.Name == key).Select(s => new SectionModel(s));
            }
        }

        public static string NormalizeMetaKey(string key)
        {
            return key.Replace("-", "_");
        }
    }

    public class SectionModel
    {
        private Section section;
        public ExpandoObject Meta { get; private set; }
        public ExpandoObject Metas { get; private set; }

        public SectionMap Map { get; private set; }

        public Dictionary<string, ExpandoObject> Data { get; private set; }

        public string Content => section.Content;
        public string Name => section.Name;

        public int WordCount => section.WordCount;

        public ExpandoObject Sections { get; private set; }
        public ExpandoObject Section { get; private set; }

        public IEnumerable<string> Lines => section.Content.Split(Environment.NewLine);

        public SectionModel(Section section)
        {
            this.section = section;
            Map = section.Map;

            Meta = new ExpandoObject();
            Metas = new ExpandoObject();
            //Data = new Dictionary<string, ExpandoObject>();

            //if (section.Meta.Any(m => m.Key.StartsWith("data-")))
            //{
            //    foreach (var dataMeta in section.Meta.Where(m => m.Key.StartsWith("data-")))
            //    {
            //        var keyName = dataMeta.Key.Split('-').Last();

            //        var c = new HttpClient();
            //        var data = JsonConvert.DeserializeObject<ExpandoObject>(c.GetStringAsync(dataMeta.Value.ToString()).Result, new ExpandoObjectConverter());
            //        Data.Add(keyName, data);
            //    }
            //}

            foreach (var meta in section.Meta.GroupBy(m => m.Key))
            {
                string key, value;
                List<object> values;

                key = meta.Key;
                value = meta.FirstOrDefault()?.Value?.ToString();
                values = meta.Select(m => m.Value).ToList();

                // If the value has a length and all characters are numbers, then treat it as a number...
                ((IDictionary<string, object>)Meta)[DocumentModel<KiwiDocument>.NormalizeMetaKey(key)] = value != null && value.Length > 0 && value.All(c => char.IsDigit(c)) ? (object)long.Parse(value) : (object)value;
                ((IDictionary<string, object>)Metas)[DocumentModel<KiwiDocument>.NormalizeMetaKey(key)] = values;
            }

            Sections = new ExpandoObject();
            ((IDictionary<string, object>)Sections)["all"] = section.Sections.Select(s => new SectionModel(s));
            Section = new ExpandoObject();
            foreach (var distinctSection in section.Sections.GroupBy(s => s.Name))
            {
                ((IDictionary<string, object>)Sections)[NormalizeMetaKey(distinctSection.Key)] = distinctSection.Select(s => new SectionModel(s));
                ((IDictionary<string, object>)Section)[NormalizeMetaKey(distinctSection.Key)] = new SectionModel(distinctSection.First());
            }

        }

        public static string NormalizeMetaKey(string key)
        {
            return key.Replace("-", "_");
        }

    }
}
