﻿using DeaneBarker.Kiwi.Parsing;
using System;
using System.Collections.Generic;
using System.Linq;

namespace DeaneBarker.Kiwi
{
    public class KiwiDocument : StructuredTextBase<KiwiDocument>
    {
        private readonly string content; // This will keep the raw content forever

        // Meta
        public override MetaCollection Meta => Sections.FirstOrDefault()?.Meta ?? new MetaCollection();
        public MetaCollection AllMeta => new MetaCollection(Sections.SelectMany(s => s.Meta));


        // Named Meta
        public string Title => Meta.Get<string>("title");
        public string Status => Meta.Get<string>("status");
        public string Description => Meta.Get<string>("description") ?? Meta.Get<string>("desc");
        public DateTime Date => DateTime.TryParse(Meta.Get<string>("date") ?? string.Empty, out DateTime parsedDate) ? parsedDate : DateTime.MinValue;
        public bool HasDate => Date != DateTime.MinValue;


        // Sections and Content
        public List<Section> NarrativeSections => FilteredSections.Where(s => s.LeftOpen).ToList();
        public string NarrativeContent => string.Join(Environment.NewLine, NarrativeSections.SelectMany(s => s.TrimmedLines));
        public List<Section> DefaultSections => Sections.Where(s => s.Name == Settings.Instance.DefaultSectionName).ToList();
        public string DefaultContent => string.Join(Environment.NewLine, DefaultSections.SelectMany(s => s.TrimmedLines));
        public string Content => Sections.FirstOrDefault()?.Content;

        // This is just for storing general strongly-typed data
        // It cannot be accessed except by code -- Kiwi source will never write to it
        // It's handy for when a process has to store something to be processed later in a pipeline
        // Seldom-used, but very handy when you need it
        public Dictionary<string, object> Items { get; } = new();


        public KiwiDocument(string input = null, bool processIncludes = true)
        {
            // Just in case we need it for anything...
            content = input;

            if (string.IsNullOrWhiteSpace(input)) return;

            var sectionSources = SourceParser.GetSectionSources(input);
            foreach (var sectionSource in sectionSources)
            {
                // This gets a section based on nothing but the source
                // It may not have a name when we get it back, it depends on if a name was in the source
                // The the name is empty, we'll name it before adding it
                var section = SectionParser.ParseSection(sectionSource);

                // If no name, then name it after (1) the last section left open, or (2) the default
                section.Name ??= Sections.Where(s => s.LeftOpen).LastOrDefault()?.Name ?? Settings.Instance.DefaultSectionName;

                // Place the section based on the naming
                if (section.Name.StartsWith('>'))
                {
                    section.Name = section.Name.Trim('>');
                    Sections.Last().Sections.Add(section);
                }
                else
                {
                    Sections.Add(section);
                }
            }

            // Now go find sections that start _meta and elevate them
            foreach(var section in Sections)
            {
                if(section.Name.StartsWith(Settings.Instance.MetaSectionPrefix + ':'))
                {
                    AddMeta(section.Name.Split(':').Last(), section.Content);
                }
            }
            Sections.RemoveAll(s => s.Name.StartsWith(Settings.Instance.MetaSectionPrefix + ':'));  // Should we get rid of them after elevation? I think so.

            // Process includes
            if (processIncludes)
            {
                while (HasSection(Settings.Instance.IncludeSectionName))
                {
                    var section = Sections.First(s => s.Name == Settings.Instance.IncludeSectionName);
                    var index = Sections.FindIndex(s => s.Name == Settings.Instance.IncludeSectionName);
                    Sections.Remove(section);

                    if (Settings.Instance.IncludeProvider == null)
                    {
                        continue;
                        //throw new InvalidOperationException("IncludeProvider has not been configured");
                    }

                    var sectionsToInclude = Settings.Instance.IncludeProvider(section.Meta);
                    Sections.InsertRange(index, sectionsToInclude.Where(s => s != null).Select(s => s.AddMeta("included", true)));
                }
            }


            // Now, go back and populate the map for each
            for (int index = 0; index <= FilteredSections.Count - 1; index++)
            {
                FilteredSections[index].Map = new SectionMap(FilteredSections, index);
            }

        }

        public static string GetMetaValue(string input, string key)
        {
            var doc = new KiwiDocument(input);
            return doc.AllMeta.Get(key);
        }

        public static string RemoveMetaValue(string input, string key)
        {
            var doc = new KiwiDocument(input);
            if (doc.Meta.HasMeta(key))
            {
                doc.Meta.Remove(doc.Meta.First(m => m.Key == "path"));
            }
            return doc.ToString();
        }

        public static string AddMetaValue(string input, string key, object value)
        {
            var doc = new KiwiDocument(input);
            doc.AddMeta(key, value);
            return doc.ToString();
        }

    }
}
