﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;

namespace DeaneBarker.Kiwi
{
    public class MetaCollection : List<MetaItem>
    {
        public Dictionary<string, object> Items => this.ToDictionary(i => i.Key, i => i.Value);
        public List<string> Keys => this.Select(i => i.Key).ToList();

        public MetaCollection()
        {
        }

        public MetaCollection(IEnumerable<MetaItem> items)
        {
            AddRange(items);
        }

        public override string ToString()
        {
            var sb = new StringBuilder();
            if (this.Any())
            {
                if (Settings.Instance.SerializationMode == SerializationMode.Verbose)
                {
                    sb.AppendLine(Settings.Instance.FrontMatterDelim);
                }
                foreach (var item in this)
                {
                    sb.AppendLine($"{item.Key}{Settings.Instance.MetaDelim} {item.Value}");
                }
                if (Settings.Instance.SerializationMode == SerializationMode.Verbose)
                {
                    sb.AppendLine(Settings.Instance.FrontMatterDelim);
                }
            }
            sb.Append(Settings.Instance.PostMetaBlockDelim);
            return sb.ToString();
        }

        public void Add(string owner, string key, object value)
        {
            key = NormalizeKey(key);
            base.Add(new MetaItem(owner, key, value));

        }

        public bool ContainsKey(string key)
        {
            key = NormalizeKey(key);
            return this.Any(m => m.Key == key);
        }

        public object this[string key]
        {
            get
            {
                key = NormalizeKey(key);
                return this.FirstOrDefault(m => m.Key == key).Value;
            }
        }

        public string FindMetaKey(string keyString)
        {
            keyString = NormalizeKey(keyString);
            foreach (var key in keyString.Split(Settings.Instance.MetaKeyDelim.ToCharArray()).Select(s => s.Trim()))
            {
                if (HasMeta(key))
                {
                    return key;
                }
            }
            return null;
        }

        public bool HasMeta(string key)
        {
            key = NormalizeKey(key);
            return this.ContainsKey(key);
        }

        public object GetString(string key, string defaultValue = "")
        {
            key = NormalizeKey(key);
            key = FindMetaKey(key);
            if(key == null)
            {
                return defaultValue;
            }
            return Get<string>(key);
        }

        public List<string> GetStrings(string key)
        {
            key = NormalizeKey(key);
            key = FindMetaKey(key);
            if (key == null)
            {
                return new List<string>();
            }
            return this.Where(m => m.Key == key).Select(m => m.Value?.ToString()).ToList();
        }

        public string Get(string key)
        {
            var keys = key.Split(',').Select(s => s.Trim());

            foreach(var thisKey in keys)
            {
                var value = Get<String>(thisKey);
                if(value != null)
                {
                    return value;
                }
            }

            return null;
        }

        public T Get<T>(string key)
        {
            key = NormalizeKey(key);
            key = FindMetaKey(key);
            if (key == null)
            {
                return default(T);
            }

            var value = this.FirstOrDefault(m => m.Key == key)?.Value;

            if (typeof(T) == typeof(DateTime))
            {
                if(DateTime.TryParse(value.ToString(), out DateTime parsedDate))
                {
                    value = parsedDate;
                }
            }

            if(typeof(T) == typeof(int))
            {
                if(!int.TryParse(value.ToString(), out _))
                {
                    throw new Exception($"Value of {value} cannot be parsed as int");
                }
            }

            return (T)value;
        }



        [Obsolete("Use null coalesce instead")]
        public T GetFirstWithValue<T>(params string[] keys)
        {
            foreach(var key in keys)
            {
                var value = GetAllValues<T>(key);
                if (value.Any())
                {
                    return value.FirstOrDefault();
                }
            }
            return default(T);
        }

        [Obsolete("Get<string> will do the same thing")]
        public T GetFirstOfKey<T>(string key)
        {
            return GetAllValues<T>(key).FirstOrDefault();
        }


        public List<T> GetAllValues<T>(string key)
        {
            key = NormalizeKey(key);
            key = FindMetaKey(key);
            if(key == null)
            {
                return new List<T>();
            }
            return this.Where(m => m.Key == key).Select(m => m.Value).Cast<T>().ToList();
        }

        public static string NormalizeKey(string key)
        {
            return key.ToLower();
        }

        public void Populate(string sectionName, Lines lines)
        {
            // If we have no lines, skip it...
            if(lines == null || !lines.Any())
            {
                return;
            }

            // 1. Verbose method
            if(lines.First() == Settings.Instance.FrontMatterDelim)
            {
                lines.GetNext();  // Skip the first line
                while(true)
                {
                    var thisLine = lines.GetNext();
                    if(thisLine == Settings.Instance.FrontMatterDelim || thisLine == null)
                    {
                        lines.GetNext(); // Skip the last line (if there is one)
                        break;
                    }
                    AddMetaFromLine(thisLine);
                }
                lines.RemoveBeforeCurrent(); // We have the meta, so throw all that away
                return; // If we have verbose, we don't check for concise
            }

            // 2. Concise method
            if(
                !string.IsNullOrWhiteSpace(lines.First()) &&                                                            // 1. The first line isn't blank
                lines.First().Split(' ').First().Contains(Settings.Instance.MetaDelim) &&                               // 2.The first line has a colon before any space
                lines.TakeWhile(s => !String.IsNullOrWhiteSpace(s)).All(s => s.Contains(Settings.Instance.MetaDelim))   // 3. Every line has a colon before the next blank line
            )   
            {
                while (true)
                {
                    var thisLine = lines.GetNext();
                    if (String.IsNullOrWhiteSpace(thisLine) || !thisLine.Contains(Settings.Instance.MetaDelim))
                    {
                        break;
                    }
                    AddMetaFromLine(lines.Current);
                }
                lines.RemoveBeforeCurrent(); // We have the meta, so throw all that away
            }

            // Internal function
            void AddMetaFromLine(string line)
            {
                var parts = line.Split(Settings.Instance.MetaDelim.ToCharArray());
                Add(sectionName, parts.First().Trim(), line.Substring(line.IndexOf(Settings.Instance.MetaDelim) + 1).Trim());
            }
        }

        public XElement ToXml()
        {
            return new XElement("meta",
                this.Select(m => new XElement(m.Key, m.Value.ToString()))
                );
        }
    }
}
