﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace DeaneBarker.Kiwi
{
    [DebuggerDisplay("{Name.ToUpper()}, Meta: {Meta.Count}, Lines: {TrimmedLines.Count}")]
    public class Section : StructuredTextBase<Section>
    {
        public int SourceLineNumber { get; init; }

        public KiwiDocument OwningDocument { get; private set; }
        public string Name { get; set; }

        private StringBuilder content = new();
        public string Content
        {
            get { return content.ToString(); }
            set { content = new StringBuilder(value); }
        }
        public SectionMap Map { get; set; }
        public bool IsEmpty => string.IsNullOrWhiteSpace(Content) && !Meta.Any();
        public override MetaCollection Meta { get; } = new();

        private Lines lines;
        public ReadOnlyCollection<string> Lines => lines.AsReadOnly();
        public bool LeftOpen { get; set; } = true;
        public List<string> TrimmedLines => lines.SkipWhile(l => string.IsNullOrWhiteSpace(l)).Reverse().SkipWhile(l => string.IsNullOrWhiteSpace(l)).Reverse().ToList();

        public Section(string name)
        {
            Sections = new List<Section>();
        }

        public bool IsExcluded()
        {
            return Regex.IsMatch(Name, Settings.Instance.ExcludePattern);
        }

        public new Section AddMeta(string key, object value)
        {
            Meta.Add(new MetaItem(Name, key, value));
            return this;
        }

        public Section(string name, string content, KiwiDocument owningDocument = null, int sourceLineNumber = 0)
        {
            Sections = new List<Section>();
            Name = name;
            OwningDocument = owningDocument;
            Meta = new MetaCollection();
            lines = new Lines(content);
            Meta.Populate(name, lines);
            SourceLineNumber = sourceLineNumber;

            // We need to have a genuine NULL if there's no content
            var contentString = lines.ToStringRemaining().Trim();
            Content = string.IsNullOrWhiteSpace(contentString) ? null : contentString;
        }

        public override string ToString()
        {
            var output = new StringBuilder()
                .Append(Name != Settings.Instance.DefaultSectionName || Map.Index > 0 ? $"{Settings.Instance.SectionStart} {Name} {Settings.Instance.SectionEnd}" : Settings.Instance.PostSectionHeaderDelim)
                .Append(Environment.NewLine)
                .Append(Meta.ToString())
                .Append(Content)
                .ToString();
            return output;
        }

        public void AppendNewLine()
        {
            AppendContent(Environment.NewLine);
        }

        public void AppendContent(string contentToAppend)
        {
            content.Append(contentToAppend);
        }
            

    }
}
