﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml.Linq;

namespace DeaneBarker.Kiwi
{
    public abstract class StructuredTextBase<T> where T : StructuredTextBase<T>
    {
        public abstract MetaCollection Meta { get; }
        public List<Section> Sections { get; set; } = new();
        public List<Section> FilteredSections => Sections.Where(s => !s.IsExcluded()).ToList();
        public int SectionCount => Sections.Count();
        public int WordCount => FilteredSections.Sum(s => s.WordCount);

        public T AddSection(string name, string input)
        {
            name ??= Settings.Instance.DefaultSectionName;
            Sections.Add(new Section(name, input));
            return (T)this;
        }

        public T AddContent(string content)
        {
            if (Sections.Any())
            {
                Sections.First().Content = content;
            }
            else
            {
                Sections.Add(new Section(Settings.Instance.DefaultSectionName, content));
            }
            return (T)this;
        }

        public T AddMeta(string key, object value)
        {
            return AddMeta(key, new[] { value });
        }

        public T AddMeta(string key, IEnumerable<object> values)
        {
            if (!Sections.Any())
            {
                AddSection(Settings.Instance.DefaultSectionName, string.Empty);
            }

            values.ToList().ForEach(v =>
            {
                Sections.First().Meta.Add(Sections.First().Name, key, v);
            });
            return (T)this;
        }

        public MetaCollection GetSectionMeta(string key)
        {
            return Sections.FirstOrDefault(s => s.Name == key)?.Meta;
        }

        public bool HasSection(string key)
        {
            return Sections.Any(s => s.Name == key);
        }

        public Section GetSection(string key)
        {
            return Sections.FirstOrDefault(s => s.Name == key);
        }

        public Section GetSection(IEnumerable<string> keys)
        {
            return Sections.FirstOrDefault(s => keys.Contains(s.Name));
        }

        public string GetSectionContent(string key, string defaultValue = null)
        {
            return Sections.FirstOrDefault(s => s.Name == key)?.Content ?? defaultValue;
        }

        public IEnumerable<Section> GetAllSections(params string[] keys)
        {
            return FilterSectionSet(Sections, keys);
        }

        public IEnumerable<Section> GetSections(params string[] keys)
        {
            return FilterSectionSet(FilteredSections, keys);
        }

        public IEnumerable<Section> FilterSectionSet(IEnumerable<Section> workingSet, params string[] keys)
        {
            var exclusionIndicator = '-';
            var delimiter = ',';

            // No keys means they get back all the narrative sections
            if (!keys.Any())
            {
                return workingSet;
            }

            // Split any keys on delimiter
            keys = keys.SelectMany(k => k.Split(delimiter)).Select(s => s.ToLower().Trim()).ToArray();

            var inclusions = keys.Where(k => !k.StartsWith(exclusionIndicator));
            var exclusions = keys.Where(k => k.StartsWith(exclusionIndicator)).Select(k => k.TrimStart(exclusionIndicator));

            // Start off with just the inclusions
            if (inclusions.Any())
            {
                workingSet = workingSet.Where(s => inclusions.Contains(s.Name));
            }

            // Filter out the exclusions
            if (exclusions.Any())
            {
                workingSet = workingSet.Where(s => !exclusions.Contains(s.Name));
            }

            return workingSet;
        }

        public void Apply(Func<Section, StructuredTextBase<T>, Section> function)
        {
            Sections = Sections.Select(s => s.IsExcluded() ? s : function(s, this)).ToList();
        }

        public void Apply(Func<string, string> function)
        {
            Sections.ToList().ForEach(s =>
            {
                if (!s.IsExcluded())
                {
                    s.Content = function(s.Content);
                }
            });
        }

        public override string ToString()
        {
            return string.Join(Settings.Instance.SectionSeparator, Sections).Trim();
        }

        public XDocument ToXml()
        {
            return new XDocument(
                new XElement("document",
                    Sections.Select(s => s.ToXml())
                    )
               );
        }

        public Dictionary<string, string> ToDictionary()
        {
            return FilteredSections
                .Concat(Meta.Select(m => new Section(m.Key, m.Value.ToString())))
                .ToDictionary(s => s.Name, s => s.Content);
        }

        public Dictionary<string, object> ToObjectDictionary()
        {
            return FilteredSections
                .Concat(Meta.Select(m => new Section(m.Key, m.Value.ToString())))
                .ToDictionary(s => s.Name, s => (object)s.Content);
        }

        public void ApplyStructure(string structureSpecification)
        {
            const char wildcard = '*';
            const char delimiter = '*';

            if (string.IsNullOrWhiteSpace(structureSpecification))
            {
                // No order string
                return;
            }

            if(structureSpecification.Trim().Replace(delimiter.ToString(), string.Empty) == wildcard.ToString())
            {
                // Just a wildcard
                return;
            }

            // Note: when applying structure, hidden sections will ALWAYS go to the end
            var hiddenSections = Sections.Where(i => i.IsExcluded());

            if (!structureSpecification.Any(c => c == wildcard))
            {
                // No wildcard
                var orderedSections = getFromStructure(Sections, getStructureTokens(structureSpecification));
                Sections = orderedSections.Concat(hiddenSections).ToList();
            }
            else
            {
                // Wildcard, so we have to get "pre" and "post"
                var preTokens = getStructureTokens(structureSpecification.Split(wildcard).First());
                var postTokens = getStructureTokens(structureSpecification.Split(wildcard).Last());

                var preSections = getFromStructure(Sections, preTokens);
                var postSections = getFromStructure(Sections, postTokens);
                var midSections = Sections.Where(i => !i.IsExcluded() && !preTokens.Contains(i.Name) && !postTokens.Contains(i.Name));

                Sections = preSections.Concat(midSections).Concat(postSections).Concat(hiddenSections).ToList();
            }

            List<Section> getFromStructure(IEnumerable<Section> Sections, IEnumerable<string> structureTokens)
            {
                var returnValue = new List<Section>();
                foreach (var token in structureTokens)
                {
                    returnValue.AddRange(Sections.Where(s => s.Name == token));
                }

                return returnValue;
            }

            List<string> getStructureTokens(string orderString)
            {
                if (string.IsNullOrWhiteSpace(orderString)) return new List<string>();
                return orderString.Split(',').Select(c => c.Trim().ToLower()).Where(c => !string.IsNullOrWhiteSpace(c)).ToList();
            }
        }

    }
}
